import zeroneura.utils.Tensors as ts
import unittest

class TestMLP(unittest.TestCase):
    def testMLP(self):
        tensor = ts.createTensor([5, 4, 2])
        tensor = ts.applyLambdaToTensor(tensor, lambda x: round(x))

        print("Forma del tensor => ", ts.getShapeTensor(tensor))
        print(tensor)

if __name__ == "__main__":
    unittest.main()


from zeroneura.core.RecurentNeuralNetwork import RNNAFDebug as RNN
import zeroneura.utils.WordsProcessing as dp
import unittest

class TestMLP(unittest.TestCase):
    def testRNN(self):
        print(f"\n{'-'*40}Prueba Unitaria Para RecurentzeroneuralNetwork{'-'*40}\n")

        text = "hello world my name is Asher me la turbo mastuba super cabron piruje. hello world my name is Asher las ballenbas this van en el mar no en mini falda. hello world my name is Asher ollaaaaa. hello world my name is Asher coño this."

        words = text.replace("\n", " ").split()
        vocab = dp.tokenizeVocabulary(text)
        wordToIndex = dp.wordToIndex(vocab)
        indexToWord = dp.indexToWord(vocab)
        vocabSize = len(vocab)

        seqLength = 3
        inputSequences = []
        outputSequences = []

        for i in range(len(words) - seqLength):
            seqIn = words[i : i+seqLength]
            seqOut = words[i+1 : i+seqLength+1]
            inputSequences.append([dp.oneHotWord(wordToIndex[w], vocabSize) for w in seqIn])
            outputSequences.append([dp.oneHotWord(wordToIndex[w], vocabSize) for w in seqOut])

        rnn = RNN([vocabSize, 50, vocabSize], 0.1)
        rnn.train(inputSequences, outputSequences, 200, True)

        def generate_text_words(model, seed_words, length=20, temperature=1.0):
            current = seed_words[:]
            result = current[:]
            for _ in range(length):
                x = [dp.oneHotWord(wordToIndex[w], vocabSize) for w in current]
                pred = model.predict([x], logs=False)[0]
                idx = dp.sample(pred, temperature)
                next_word = indexToWord[idx]
                result.append(next_word)
                current = current[1:] + [next_word]
            return ' '.join(result)

        seed = ["hello", "world", "this"]
        print(generate_text_words(rnn, seed, length=15, temperature=0.7))

if __name__ == "__main__":
    unittest.main()


import zeroneura.utils.ActivationFunctions as af
import zeroneura.utils.Outputs as ot
import zeroneura.utils.Files as f
import random

class MLP:
    def __init__(self, structure, learningRate):
        assert isinstance(structure, list) and len(structure) >= 3, "You must provide a 3 item list in the zeroneural network structure"

        self.structure = structure
        self.learningRate = learningRate
        self.layersCount = len(structure)
        self.weights = [self.initWeights(nOut, nIn) for nIn, nOut in zip(structure[:-1], structure[1:])]
        self.biases = [[random.uniform(-1, 1) for _ in range(n)] for n in structure[1:]]
    
    def initWeights(self, rows, cols):
        return [[random.uniform(-1, 1) for _ in range(cols)] for _ in range(rows)]
    
    def feedforward(self, inputVector):
        activations = [inputVector]
        zs =[]

        for i in range(self.layersCount - 1):
            z = [sum(w * a for w, a in zip(neuron, activations[-1])) + b for neuron, b in zip(self.weights[i], self.biases[i])]
            zs.append(z)
            activations.append([af.sigmoid(val) for val in z])

        return activations, zs
    
    def printData(self):
        print(f"\n{'-'*40}Data{'-'*40}\n")
        print("Activation Function: Sigmoid")
        print("Counts:")
        print(f"Inputs: {self.structure[0]}\tHidden: {[self.structure[h] for h in range(1, self.layersCount-1)]}\tOutput: {self.structure[-1]}")
        print(f"Structure: {self.structure}\tLayers: {len(self.structure)}")
        print(f"\nWeights:\n{self.weights}")
        print(f"\nBias:\n{self.biases}")
    
    def train(self, inputs, outputs, rounds, logs = False):
        assert isinstance(inputs, (list)), "Inputs must be in a list"
        assert isinstance(outputs, (list)), "Outputs must be in a list"
        assert not (hasattr(outputs[0], '__iter__') and len(outputs[0]) != self.structure[-1]), f"Each output must have {self.structure[-1]} values to match the number of output neurons"
        

        outputs = [[y] if not isinstance(y, list) else y for y in outputs]

        for round in range(rounds):
            totalError = 0

            if logs and (round+1) % 1000 == 0:
                print(f"\n{'-'*40}Round #{round+1}{'-'*40}\n")
            
            for input, output in zip(inputs, outputs):
                activations, zs = self.feedforward(input)

                deltas = [None] * (self.layersCount - 1)
                deltas[-1] = [(yI - aI) * af.sigmoidDeriv(zI) for yI, aI, zI in zip(output, activations[-1], zs[-1])]

                for l in range(self.layersCount - 3, -1, -1):
                    layerWeights = self.weights[l + 1]
                    deltaNext = deltas[l + 1]
                    z = zs[l]
                    sp = [af.sigmoidDeriv(zI) for zI in z]
                    deltas[l] = [sum(layerWeights[k][j] * deltaNext[k] for k in range(len(deltaNext))) * sp[j] for j in range(len(sp))]

                for l in range(self.layersCount - 1):
                    for i in range(len(self.weights[l])):
                        for j in range(len(self.weights[l][i])):
                            self.weights[l][i][j] += self.learningRate * deltas[l][i] * activations[l][j]
                        self.biases[l][i] += self.learningRate * deltas[l][i]

                totalError += sum((yI - aI) ** 2 for yI, aI in zip(output, activations[-1]))

                if logs and (round + 1) % 1000 == 0:
                    print(f"Value: {input} =>\n{self.predict(input, False)}")
                    print(f"Error: {totalError}\n")

    def predict(self, inputs, logs = True):
        results = []

        inputs = [[i] if not isinstance(i, list) else i for i in inputs]
        for input in inputs:
            r = [round(o) for o in (self.feedforward(input)[0][-1])]
            results.append(r)

            if logs: print(f"{input} => {r}")

        return results

    def predictLambda(self, inputs, funcion = lambda x: x, logs = True):
        results = []

        inputs = [[i] if not isinstance(i, list) else i for i in inputs]
        for input in inputs:
            r = [funcion(o) for o in (self.feedforward(input)[0][-1])]
            results.append(r)

            if logs: print(f"{input} => {r}")

        return results

class MLPAF:
    def __init__(self, structure, learningRate, activationFunction = af.SIGMOID):
        assert isinstance(structure, list) and len(structure) >= 3, "You must provide a 3 item list in the zeroneural network structure"

        self.structure = structure
        self.learningRate = learningRate
        self.layersCount = len(structure)
        self.weights = [self.initWeights(n_out, n_in) for n_in, n_out in zip(structure[:-1], structure[1:])]
        self.biases = [[random.uniform(-1, 1) for _ in range(n)] for n in structure[1:]]
        self.activationFunction = activationFunction
        self.lastError = 0
    
    def initWeights(self, rows, cols):
        return [[random.uniform(-1, 1) for _ in range(cols)] for _ in range(rows)]
    
    def feedforward(self, inputVector):
        activations = [inputVector]
        zs =[]

        for i in range(self.layersCount - 1):
            z = [sum(w * a for w, a in zip(neuron, activations[-1])) + b for neuron, b in zip(self.weights[i], self.biases[i])]
            zs.append(z)           
            activations.append([af.activate(val, self.activationFunction) for val in z])    

        return activations, zs
    
    def train(self, inputs, outputs, rounds, logs = False):
        assert isinstance(inputs, (list)), "Inputs must be in a list"
        assert isinstance(outputs, (list)), "Outputs must be in a list"
        assert not (hasattr(outputs[0], '__iter__') and len(outputs[0]) != self.structure[-1]), f"Each output must have {self.structure[-1]} values to match the number of output neurons"

        outputs = [[y] if not isinstance(y, list) else y for y in outputs]

        for round in range(rounds):
            totalError = 0

            if logs and (round+1) % 1000 == 0:
                print(f"\n{'-'*40}Round #{round+1}{'-'*40}\n")
            
            for input, output in zip(inputs, outputs):
                activations, zs = self.feedforward(input)
                deltas = [None] * (self.layersCount - 1)

                deltas[-1] = [(yI - aI) * af.sigmoidDeriv(zI) for yI, aI, zI in zip(output, activations[-1], zs[-1])]

                for l in range(self.layersCount - 3, -1, -1):
                    layerWeights = self.weights[l + 1]
                    deltaNext = deltas[l + 1]
                    z = zs[l]
                    sp = [af.deriv(zI, self.activationFunction) for zI in z]

                    deltas[l] = [sum(layerWeights[k][j] * deltaNext[k] for k in range(len(deltaNext))) * sp[j] for j in range(len(sp))]

                for l in range(self.layersCount - 1):
                    for i in range(len(self.weights[l])):
                        for j in range(len(self.weights[l][i])):
                            self.weights[l][i][j] += self.learningRate * deltas[l][i] * activations[l][j]
                        self.biases[l][i] += self.learningRate * deltas[l][i]

                totalError += sum((yI - aI) ** 2 for yI, aI in zip(output, activations[-1]))

                if logs and (round + 1) % 1000 == 0:
                    print(f"Value: {input} =>\n{self.predict(input, False)}")
                    print(f"Error: {totalError}\n")
                
            self.lastError = totalError
    
    def printData(self):
        print(f"\n{'-'*40}Data{'-'*40}\n")
        print(f'Activation Function: {["Sigmoid", "ReLU", "Leaky ReLU", "tanh"][self.activationFunction]}')
        print("Counts:")
        print(f"Inputs: {self.structure[0]}\tHidden: {[self.structure[h] for h in range(1, self.layersCount-1)]}\tOutput: {self.structure[-1]}")
        print(f"Structure: {self.structure}\tLayers: {len(self.structure)}")
        print(f"Last Error: {self.lastError}")
        print(f"\nWeights:\n{self.weights}")
        print(f"\nBias:\n{self.biases}")

    def predict(self, inputs, logs = True):
        results = []

        inputs = [[i] if not isinstance(i, list) else i for i in inputs]
        for input in inputs:
            r = [self.feedforward(input)[0][-1]]
            results.append(r)

            if logs: print(f"{input} => {r}")

        return results

    def predictLambda(self, inputs, funcion = lambda x: x, logs = True):
        results = []

        inputs = [[i] if not isinstance(i, list) else i for i in inputs]
        for input in inputs:
            r = [funcion(o) for o in (self.feedforward(input)[0][-1])]
            results.append(r)

            if logs: print(f"{input} => {r}")

        return results
    
class MLPAFDebug:
    def __init__(self, structure, learningRate, activationFunction = af.SIGMOID):
        assert isinstance(structure, list) and len(structure) >= 3, "You must provide a 3 item list in the zeroneural network structure"

        self.structure = structure
        self.learningRate = learningRate
        self.layersCount = len(structure)
        self.weights = [self.initWeights(n_out, n_in) for n_in, n_out in zip(structure[:-1], structure[1:])]
        self.biases = [[random.uniform(-1, 1) for _ in range(n)] for n in structure[1:]]
        self.activationFunction = activationFunction
        self.lastError = 0
    
    def initWeights(self, rows, cols):
        return [[random.uniform(-1, 1) for _ in range(cols)] for _ in range(rows)]
    
    def feedforward(self, inputVector):
        activations = [inputVector]
        zs =[]

        for i in range(self.layersCount - 1):
            z = [sum(w * a for w, a in zip(neuron, activations[-1])) + b for neuron, b in zip(self.weights[i], self.biases[i])]
            zs.append(z)
            activations.append([af.activate(val, self.activationFunction) for val in z])

        return activations, zs
    
    def train(self, inputs, outputs, rounds, logs = False):
        assert isinstance(inputs, (list)), "Inputs must be in a list"
        assert isinstance(outputs, (list)), "Outputs must be in a list"
        assert not (hasattr(outputs[0], '__iter__') and len(outputs[0]) != self.structure[-1]), f"Each output must have {self.structure[-1]} values to match the number of output neurons"

        outputs = [[y] if not isinstance(y, list) else y for y in outputs]

        for round in range(rounds):
            totalError = 0

            if logs:
                progress = (round + 1) / rounds
                bar = ot.BarLoad()
                bar.showBar(progress, 'Entrenamiento: ', f' {round+1}/{rounds} ')

                if progress == 1: print()
            
            for input, output in zip(inputs, outputs):
                activations, zs = self.feedforward(input)
                deltas = [None] * (self.layersCount - 1)
                deltas[-1] = [(yI - aI) * af.sigmoidDeriv(zI) for yI, aI, zI in zip(output, activations[-1], zs[-1])]

                for l in range(self.layersCount - 3, -1, -1):
                    layerWeights = self.weights[l + 1]
                    deltaNext = deltas[l + 1]
                    z = zs[l]
                    sp = [af.deriv(zI, self.activationFunction) for zI in z]                       
                    deltas[l] = [sum(layerWeights[k][j] * deltaNext[k] for k in range(len(deltaNext))) * sp[j] for j in range(len(sp))]

                for l in range(self.layersCount - 1):
                    for i in range(len(self.weights[l])):
                        for j in range(len(self.weights[l][i])):
                            self.weights[l][i][j] += self.learningRate * deltas[l][i] * activations[l][j]
                        self.biases[l][i] += self.learningRate * deltas[l][i]

                totalError += sum((yI - aI) ** 2 for yI, aI in zip(output, activations[-1]))
                
            self.lastError = totalError
    
    def printData(self):
        print(f"\n{'-'*40}Data{'-'*40}\n")
        print(f'Activation Function: {["Sigmoid", "ReLU", "Leaky ReLU", "tanh"][self.activationFunction]}')
        print("Counts:")
        print(f"Inputs: {self.structure[0]}\tHidden: {[self.structure[h] for h in range(1, self.layersCount-1)]}\tOutput: {self.structure[-1]}")
        print(f"Structure: {self.structure}\tLayers: {len(self.structure)}")
        print(f"Last Error: {self.lastError}")
        print(f"\nWeights:\n{self.weights}")
        print(f"\nBiases:\n{self.biases}")

    def saveData(self, outputName = 'zeroneuralNetworkData', outputPath = './'):
        DATA = f.getMLPDATA()
        DATA["LayersCount"] = self.layersCount
        DATA["Structure"] = self.structure
        DATA["LearningRate"] = self.learningRate
        DATA["ActivationFunction"] = self.activationFunction
        DATA["Weights"] = self.weights
        DATA["Biases"] = self.biases
        
        return f.saveMLPDataNND(DATA, outputName, outputPath)
    
    def loadData(self, inputName = 'zeroneuralNetworkData', inputPath = './'):
        DATA = f.loadMLPDataNND(inputName, inputPath)

        self.layersCount = DATA["LayersCount"]
        self.structure = DATA["Structure"]
        self.learningRate = DATA["LearningRate"]
        self.activationFunction = DATA["ActivationFunction"]
        self.weights = DATA["Weights"]
        self.biases = DATA["Biases"]

        return True

    def predict(self, inputs, funcion = lambda x: x, logs = True):
        results = []

        inputs = [[i] if not isinstance(i, list) else i for i in inputs]
        for input in inputs:
            r = [funcion(o) for o in (self.feedforward(input)[0][-1])]
            results.append(r)

            if logs: print(f"{input} => {r}")

        return results


import zeroneura.utils.Images as im
import unittest

class TestMLP(unittest.TestCase):
    def testMLP(self):
        inputPath = "./tests/InputsCNN/FresaT.png"
        outputPath = "./tests/InputsCNN/FresaTR.png"

        im.resize(inputPath, [150, 150], outputPath)

if __name__ == "__main__":
    unittest.main()


import ast

def getMLPDATA(): 
    return {"LayersCount": None, "Structure": None, "LearningRate": None, "ActivationFunction": None, "Weights": None, "Biases": None}

def getCNNDATA(): 
    return {"Kernels": None, "LayersCount": None, "Structure": None, "LearningRate": None, "ActivationFunction": None, "Weights": None, "Biases": None}

def loadMLPDataNND(inputName = 'zeroneuralNetworkData', inputPath = './'):
    DATA = getMLPDATA()

    with open(f"{inputPath}{inputName}.nnd", "r", encoding="utf-8") as file:
        content = file.read()

    parts = content.split("Weights:")
    head = parts[0].strip()
    wieghtsAndBiases = parts[1].split("Biases:")
    weights = wieghtsAndBiases[0].strip()
    biases = wieghtsAndBiases[1].strip()

    DATA["Weights"] = ast.literal_eval(weights)
    DATA["Biases"] = ast.literal_eval(biases)

    for line in head.splitlines():
        if line.startswith("LayersCount:"):
            DATA["LayersCount"] = int(line.split(":", 1)[1].strip())
        elif line.startswith("Structure:"):
            DATA["Structure"] = ast.literal_eval(line.split(":", 1)[1].strip())
        elif line.startswith("ActivationFunction:"):
            DATA["ActivationFunction"] = int(line.split(":", 1)[1].strip())
        elif line.startswith("LearningRate:"):
            DATA["LearningRate"] = float(line.split(":", 1)[1].strip())

    return DATA

def saveMLPDataNND(MLPDATA, outputName = 'zeroneuralNetworkData', outputPath = './'):
    with open(f"{outputPath}{outputName}.nnd", "w") as file:
        file.write("LayersCount: "+str(MLPDATA["LayersCount"])+"\n")
        file.write("Structure: "+str(MLPDATA["Structure"])+"\n")
        file.write("ActivationFunction: "+str(MLPDATA["ActivationFunction"])+"\n")
        file.write("LearningRate: "+str(MLPDATA["LearningRate"])+"\n")
        file.write("Weights: \n"+__formatTensor(MLPDATA["Weights"])+"\n")
        file.write("Biases: \n"+__formatTensor(MLPDATA["Biases"]))
        
    return True

def loadCNNDataNND(inputName = 'zeroneuralNetworkData', inputPath = './'):
    DATA = getCNNDATA()

    with open(f"{inputPath}{inputName}.nnd", "r", encoding="utf-8") as file:
        content = file.read()

    parts = content.split("Weights:")
    head = parts[0].strip()
    wieghtsAndBiases = parts[1].split("Biases:")
    weights = wieghtsAndBiases[0].strip()
    biasesAndKernels = wieghtsAndBiases[1].split("Kernels:")
    biases = biasesAndKernels[0].strip()
    kernels = biasesAndKernels[1].strip()

    DATA["Weights"] = ast.literal_eval(weights)
    DATA["Biases"] = ast.literal_eval(biases)
    DATA["Kernels"] = ast.literal_eval(kernels)

    for line in head.splitlines():
        if line.startswith("LayersCount:"):
            DATA["LayersCount"] = int(line.split(":", 1)[1].strip())
        elif line.startswith("Structure:"):
            DATA["Structure"] = ast.literal_eval(line.split(":", 1)[1].strip())
        elif line.startswith("ActivationFunction:"):
            DATA["ActivationFunction"] = int(line.split(":", 1)[1].strip())
        elif line.startswith("LearningRate:"):
            DATA["LearningRate"] = float(line.split(":", 1)[1].strip())

    return DATA

def saveCNNDataNND(CNNDATA, outputName = 'zeroneuralNetworkData', outputPath = './'):
    with open(f"{outputPath}{outputName}.nnd", "w") as file:
        file.write("LayersCount: "+str(CNNDATA["LayersCount"])+"\n")
        file.write("Structure: "+str(CNNDATA["Structure"])+"\n")
        file.write("ActivationFunction: "+str(CNNDATA["ActivationFunction"])+"\n")
        file.write("LearningRate: "+str(CNNDATA["LearningRate"])+"\n")
        file.write("Weights:\n"+__formatTensor(CNNDATA["Weights"])+"\n")
        file.write("Biases:\n"+__formatTensor(CNNDATA["Biases"])+"\n")
        file.write("Kernels:\n"+__formatTensor(CNNDATA["Kernels"]))
        
    return True

def __formatTensor(tensor, indent=1, tab="\t"):
    formatted = "[\n"
    for layer in tensor:
        if isinstance(layer[0], (list, tuple)):
            formatted += tab * indent + "[\n"
            for row in layer:
                formatted += tab * (indent + 1)
                formatted += "[" + ", \t".join(f"{v}" for v in row) + "], \n"

            formatted = formatted.rstrip(", \n") + "\n"
            formatted += tab * indent + "], \n"
        else:
            formatted += tab * indent
            formatted += "[" + ", \t".join(f"{v}" for v in layer) + "], \n"

    formatted = formatted.rstrip(", \n") + "\n]"
    return formatted


class Perceptron:
    def __init__(self, weight, rateLearning, bias):
        assert isinstance(weight, (list)), "Weights must be in a list"

        self.weight = weight
        self.rateLearning = rateLearning
        self.bias = bias

    def activate(self, sum):
        return 1 if sum > 0 else 0


    def train(self, inputs, outputs, rounds, logs = False):
        assert isinstance(inputs, (list)), "Inputs must be in a list"
        assert isinstance(outputs, (list)), "Outputs must be in a list"

        if len(inputs) == len(outputs):

            for round in range(rounds):
                if logs:
                    print(f"\n{'-'*40}Round #{round+1}{'-'*40}\n")

                for i in range(len(inputs)):
                    x = inputs[i]
                    expected = outputs[i]

                    gotten = self.activate(self.calculate(x))
                    error =  expected - gotten

                    if error != 0:
                        if hasattr(x, '__iter__') and not isinstance(x, (str, bytes)):
                            for j in range(len(x)):
                                self.weight[j] += self.rateLearning * error * x[j]
                        else:
                            self.weight[0] += self.rateLearning * error * x

                        self.bias += self.rateLearning*error

                    if logs:
                        print(f"{'-'*40}Inputs{'-'*40}")
                        print(x)
                        print(f"{'-'*40}Outputs{'-'*40}")
                        print(f"Expected: {expected}\tGotten: {gotten}")
                        print(f"Weigths: {self.weight}\tError: {error}\tBias: {self.bias}\n")

            return 1
        else:
            print("You must provide the same number of inputs and outputs")
            return 0
        
    def calculate(self, x):
        sum = 0

        if len(x) == len(self.weight):
            if hasattr(x, '__iter__') and not isinstance(x, (str, bytes)):
                for j in range(len(x)):
                    sum += x[j] * self.weight[j]
            else:
                sum += x*self.weight[0]
            
            sum += self.bias
        
        else:
            print("You must provide the same number of values of inputs and weights")
        
        return sum
    
    def printData(self):
        print(f"{'-'*40}Data{'-'*40}")
        print(f"Weigths: {self.weight}\tBias: {self.bias}\n")


    def predict(self, inputs, logs = False):
        assert isinstance(inputs, (list)), "Inputs must be in a list"

        results = []
        for x in inputs:
            r = self.activate(self.calculate(x))
            results.append(r)

            if logs:
                print(f"Inputs: {x}\t->\tResult: {r}")

        return results


import random, math

def tokenizeVocabulary(text):
    text = text.replace("\n", " ").replace("\t", " ").split()
    return sorted(list(set(text)))

def wordToIndex(words):
    return {word:index for index, word in enumerate(words)}

def indexToWord(words):
    return {index:word for index, word in enumerate(words)}

def oneHotWord(index, size):
    return [1 if i == index else 0 for i in range(size)]

def softmax(x, temperature = 1.0):
    x = [i/temperature for i in x]
    
    maxX = max(x)
    exp = [math.exp(i-maxX) for i in x]
    sumExp = sum(exp)

    return [j/sumExp for j in exp]

def sample(x, temperature = 1.0):
    probs = softmax(x, temperature)
    return random.choices(range(len(probs)), weights=probs)[0]


import zeroneura.utils.ActivationFunctions as af
import random

class RNNAFDebug:
    def __init__(self, structure, learningRate = 0.1, activationFunction = af.SIGMOID):
        assert isinstance(structure, list) and len(structure) >= 3, "Provide 3+ item list for network structure"
        self.structure = structure
        self.learningRate = learningRate
        self.layersCount = len(structure)
        self.activationFunction = activationFunction
        self.lastError = 0

        self.weights = []
        self.hiddenWeights = []
        self.biases = []

        for layerIn, layerOut in zip(structure[:-1], structure[1:]):
            self.weights.append([[random.uniform(-1,1) for _ in range(layerIn)] for _ in range(layerOut)])
            self.hiddenWeights.append([[random.uniform(-1,1) for _ in range(layerOut)] for _ in range(layerOut)])
            self.biases.append([random.uniform(-1,1) for _ in range(layerOut)])

    def feedforward(self, inputSequence):
        hiddenPrev = [[0]*n for n in self.structure[1:]]
        activationsSeq = []
        zsSeq = []

        for x in inputSequence:
            activations = [x]
            zs = []

            for l in range(self.layersCount-1):
                z = [sum(w*a for w,a in zip(neuron, activations[-1])) + sum(hw*hp for hw,hp in zip(neuronHidden, hiddenPrev[l])) + b for neuron, neuronHidden, b in zip(self.weights[l], self.hiddenWeights[l], self.biases[l])]
                zs.append(z)
                a = [af.activate(val, self.activationFunction) for val in z]
                activations.append(a)
                hiddenPrev[l] = a

            activationsSeq.append(activations)
            zsSeq.append(zs)

        return activationsSeq, zsSeq

    def train(self, inputSequences, outputSequences, rounds, logs = False):
        for round in range(rounds):
            totalError = 0

            if logs:
                progress = (round+1)/rounds
                bar_len = 50
                filled_len = int(bar_len*progress)
                bar = '█' * filled_len + '-'*(bar_len - filled_len)
                print(f'\rEntrenamiento: |{bar}| {round+1}/{rounds} ({progress*100:.2f}%)', end='', flush=True)
                if progress==1: print()

            for seqIn, seqOut in zip(inputSequences, outputSequences):
                activationsSeq, zsSeq = self.feedforward(seqIn)
                deltasSeq = [[None] * (self.layersCount-1) for _ in seqIn]

                for t in reversed(range(len(seqIn))):
                    y = seqOut[t] if isinstance(seqOut[t], list) else [seqOut[t]]
                    deltasSeq[t][-1] = [(yi - ai) * af.deriv(z, self.activationFunction) for yi, ai, z in zip(y, activationsSeq[t][-1], zsSeq[t][-1])]

                    for l in range(self.layersCount-3,-1,-1):
                        layerWeights = self.weights[l+1]
                        deltaNext = deltasSeq[t][l+1]
                        sp = [af.deriv(val, self.activationFunction) for val in zsSeq[t][l]]
                        deltasSeq[t][l] = [sum(layerWeights[k][j]*deltaNext[k] for k in range(len(deltaNext)))*sp[j] for j in range(len(sp))]

                    for l in range(self.layersCount-1):
                        for i in range(len(self.weights[l])):
                            for j in range(len(self.weights[l][i])):
                                self.weights[l][i][j] += self.learningRate * deltasSeq[t][l][i] * activationsSeq[t][l][j]
                            self.biases[l][i] += self.learningRate * deltasSeq[t][l][i]

                totalError += sum((y - a)**2 for t in range(len(seqOut)) for y,a in zip(seqOut[t], activationsSeq[t][-1]))
            self.lastError = totalError

    def printData(self):
        print(f"\n{'-'*40}Data{'-'*40}\n")
        print(f'Activation Function: {["Sigmoid","ReLU","Leaky ReLU","Tanh"][self.activationFunction]}')
        print(f"Structure: {self.structure}\tLayers: {self.layersCount}")
        print(f"Last Error: {self.lastError}")
        print(f"\nWeights:\n{self.weights}")
        print(f"\nBiases:\n{self.biases}")

    def predictLambda(self, inputSequences, funcion=lambda x:x, logs=True):
        results=[]
        for seq in inputSequences:
            act = self.feedforward(seq)[0]
            sequenceResults = [[funcion(v) for v in act_step[-1]] for act_step in act]
            results.append(sequenceResults)
            if logs: print(f"{seq} => {sequenceResults}")
        return results

    def predict(self, inputSequences, logs=True):
        results=[]
        for seq in inputSequences:
            act = self.feedforward(seq)[0]
            r = [v for v in act[-1][-1]]
            results.append(r)
            if logs: print(f"{seq} => {r}")
        return results


import math

SIGMOID = 0
RELU = 1
LEAKYRELU = 2
TANH = 3

def ReLU(x): return max(0, x)    
def ReLUDeriv(x): return 1 if x > 0 else 0

def leakyReLU(x, alpha = 0.01): return x if x > 0 else alpha * x
def leakyReLUDeriv(x, alpha = 0.01): return 1 if x > 0 else alpha

def tanh(x): return math.tanh(x)    
def tanhDeriv(x): return 1 - math.tanh(x) ** 2

def sigmoid(x): return 1 / (1 + math.exp(-x))
def sigmoidDeriv(x): s = sigmoid(x); return s * (1 - s)

def activate(x, activationFunction = SIGMOID):
    if activationFunction == RELU: return ReLU(x)
    if activationFunction == LEAKYRELU: return leakyReLU(x)
    if activationFunction == TANH: return tanh(x)
    return sigmoid(x)

def deriv(x, activationFunction = SIGMOID):
    if activationFunction == RELU: return ReLUDeriv(x)
    if activationFunction == LEAKYRELU: return leakyReLUDeriv(x)
    if activationFunction == TANH: return tanhDeriv(x)
    return sigmoidDeriv(x)


import zeroneura.utils.ActivationFunctions as af
import random

def createTensor(structure, min = 0, max = 1, function = lambda x: x):    
    if len(structure) == 1:
        return [function(random.uniform(min, max)) for _ in range(structure[0])]
    
    tensor = []
    for _ in range(structure[0]): tensor.append(createTensor(structure[1:], min, max, function))

    return tensor

def getShapeTensor(tensor):
    if isinstance(tensor, (int, float)):
        return []
    if not tensor:  # tensor vacío
        return [0]
    if isinstance(tensor[0], (int, float)):
        return [len(tensor)]
    return [len(tensor)] + getShapeTensor(tensor[0])


def applyLambdaToTensor(tensor, function = lambda x: x):
    if isinstance(tensor, (int, float)):
        return function(tensor)
    if not tensor:  # lista vacía
        return tensor
    if isinstance(tensor[0], (int, float)):
        return [function(value) for value in tensor]
    return [applyLambdaToTensor(sub, function) for sub in tensor]


import zeroneura.utils.Tensors as ts
import zeroneura.utils.DataProcessing as dp
import zeroneura.utils.ActivationFunctions as af
import zeroneura.utils.Images as im
import zeroneura.utils.Outputs as ot
import zeroneura.utils.Files as f
from zeroneura.core.MultiLayerPerceptron import MLPAFDebug as MLP
import numpy as np, cv2

class CNN:
    def __init__(self, kernels, padding = 0, stride = 1):
        self.kernels = kernels
        self.padding = padding
        self.stride = stride
        self.mlp = MLP([100, 20, 10, 1], 0.1, af.RELU)
    
    def convolve(self, img, kernel, stride = 1):
        return dp.applyKernelWithStride(img, kernel, stride)
    
    def forward(self, img, layers):
        img = im.applyGrayScale(img) if isinstance(img, str) else img

        for layerKernel in layers:
            features = []

            for kernel in layerKernel:
                fmap = self.convolve(img, kernel, self.stride)
                fmap = ts.applyLambdaToTensor(fmap, lambda x: af.ReLU(x))
                fmap = dp.maxPooling(fmap)
                features.append(fmap)

            img = features
    
        return img
    
    def train(self, frutas):
        inputs = []
        outputs = []

        for indexfruta, fruta in enumerate(frutas):
            print(f"Fruta a Analizar: {fruta}[{indexfruta+1}]")

            features = self.forward(f"./tests/Inputs/{fruta}.png")
            for _ in range(3):
                for index, feature in enumerate(features):
                    features[index] = dp.maxPooling(feature)
            CNN.showOutputs(features, f"Convolve{fruta}", f"./tests/Outputs/{fruta}/")
            print()

            for feature in features:
                feature = dp.flatten(feature)
                feature = [min(val, 255) for val in feature]
                feature = [val/255 for val in feature]
                inputs.append(feature)
                outputs.append((indexfruta+1)/10)

        self.mlp.train(inputs, outputs, 500, True)

    def predict(self, image):
        features = self.forward(image)
        for _ in range(3):
            for index, feature in enumerate(features):
                features[index] = dp.maxPooling(feature)
        CNN.showOutputs(features, f"Convolve", f"./tests/Outputs/Test/")
        print()

        maps = []

        for feature in features:
            feature = dp.flatten(feature)
            feature = [min(val, 255) for val in feature]
            feature = [val/255 for val in feature]
            maps.append(feature)

        resultados = self.mlp.predict(maps, lambda x: round(x*10), False)
        print(resultados)

        return resultados
    
    def _showOutput(img, outputName = "Output", outputPath = "./"):
        print(f"{outputName}.png")
        cv2.imwrite(f"{outputPath if outputPath != './' else ''}{outputName}.png", np.array(dp.clip(img, 0, 255)).astype(np.uint8))

    def showOutputs(featuresMaps, prefix = "Output", outputPath = "./"):
        if isinstance(featuresMaps[0], int):
            CNN._showOutput(featuresMaps, f"{prefix}", outputPath)
        else:
            for index, features in enumerate(featuresMaps):
                CNN._showOutput(features, f"{prefix}{index}", outputPath)

class CNNDebug:
    def __init__(self, kernels = [], padding = 0, stride = 1):
        self.kernels = kernels
        self.padding = padding
        self.stride = stride

        self.structureMLP = [100, 20, 10, 1]
        self.inputLayerMLP = self.structureMLP[0]
        self.mlp = MLP(self.structureMLP, 0.1, af.SIGMOID)

    def setMLPStructure(self, structureMLP, learningRate = 0.1, activationFunciton = af.SIGMOID):
        self.structureMLP = structureMLP
        self.inputLayerMLP = structureMLP[0]
        self.mlp = MLP(structureMLP, learningRate, activationFunciton)
    
    def convolve(self, img, kernel, stride = 1):
        return dp.applyKernelWithStride(img, kernel, stride)
    
    def _forwardMultiLayer(self, img, layers):      
        for layerKernel in layers:
            features = []

            for kernel in layerKernel:
                if len(ts.getShapeTensor(img)) == 2:
                    fmap = self.convolve(img, kernel, self.stride)
                else:
                    firstConv = self.convolve(img[0], kernel)
                    fmapSum = ts.createTensor(ts.getShapeTensor(firstConv), 0, 0, lambda x: round(x))

                    for subimg in img:
                        conv = self.convolve(subimg, kernel)
                        for i in range(len(conv)):
                            for j in range(len(conv[0])):
                                fmapSum[i][j] += conv[i][j]
                                
                    num_inputs = len(img)
                    for i in range(len(fmapSum)):
                        for j in range(len(fmapSum[0])):
                            fmapSum[i][j] /= num_inputs

                    fmap = fmapSum

                fmap = ts.applyLambdaToTensor(fmap, lambda x: af.ReLU(x))
                fmap = dp.maxPooling(fmap)
                features.append(fmap)


            img = features
    
        return img
    
    def forward(self, image):
        img = im.applyGrayScale(image) if isinstance(image, str) else image

        features = self._forwardMultiLayer(img, self.kernels)
        return features
    
    def train(self, imagenes, out, rounds = 1000, logs = True, outputPath = None):
        inputs = []; outputs = []

        for imagen, output in zip(imagenes, out):
            nombre = imagen.split("/")[-1].split(".")[0]
            if logs: 
                print(f"Imagen a Analizar: {nombre}")
                print(f"Nombre {nombre}")

            sizes = 0; features = imagen
            while(isinstance(features, str) or sizes > self.inputLayerMLP):
                features = self.forward(features)
                for index, feature in enumerate(features):
                    features[index] = dp.maxPooling(feature)
                
                sizes = ts.getShapeTensor(features)  
                if logs: 
                    print("Tamaño de las Imagenes", sizes, "\tDimension: ", sizes[-1]*sizes[-2])
                sizes = sizes[-1]*sizes[-2]
          
            if logs and outputPath: 
                CNNDebug.showOutputs(features, f"Convolve{nombre}", f"{outputPath}{nombre}/")
                print("\n")

            for feature in features:
                feature = dp.flatten(feature)
                feature = im.normalizeGrayImg(feature)
                
                inputs.append(self._fixLength(feature))
                outputs.append(output)

        if logs: print("\nEntrenamiento del MLP =>")
        self.mlp.train(inputs, outputs, rounds, logs)

    def predict(self, image, outputPath = None, function = lambda x: x, logs = False):
        sizes = 0; features = image
        while(isinstance(features, str) or sizes > self.inputLayerMLP):
            features = self.forward(features)
            for index, feature in enumerate(features):
                features[index] = dp.maxPooling(feature)
            
            sizes = ts.getShapeTensor(features)  
            if logs: print("Tamaño de las Imagenes", sizes)
            sizes = sizes[-1]*sizes[-2]

        if logs and outputPath:
            CNNDebug.showOutputs(features, "Convolve", outputPath)

        maps = []
        for feature in features:
            feature = dp.flatten(feature)
            feature = im.normalizeGrayImg(feature)
            maps.append(self._fixLength(feature))
        
        resultados = self.mlp.predict(maps, function, False)
        if logs: print("Resultados: ", resultados, "=> ","\n\n")

        return resultados
    
    def _fixLength(self, vec):
        target_len = self.inputLayerMLP
        vec = [float(v) for v in vec]

        if len(vec) > target_len:
            # truncar a target_len
            return vec[:target_len]
        elif len(vec) < target_len:
            return vec + ([0.0] * (target_len - len(vec)))
        else:
            return vec

    
    def saveData(self, outputName = 'zeroneuralNetworkData', outputPath = './'):
        DATA = f.getCNNDATA()
        DATA["Kernels"] = self.kernels
        DATA["LayersCount"] = self.mlp.layersCount
        DATA["Structure"] = self.mlp.structure
        DATA["LearningRate"] = self.mlp.learningRate
        DATA["ActivationFunction"] = self.mlp.activationFunction
        DATA["Weights"] = self.mlp.weights
        DATA["Biases"] = self.mlp.biases
        
        return f.saveCNNDataNND(DATA, outputName, outputPath)

    def loadData(self, inputName = 'zeroneuralNetworkData', inputPath = './'):
        DATA = f.loadCNNDataNND(inputName, inputPath)

        self.kernels = DATA["Kernels"]

        self.mlp.layersCount = DATA["LayersCount"]
        self.mlp.structure = DATA["Structure"]
        self.mlp.learningRate = DATA["LearningRate"]
        self.mlp.activationFunction = DATA["ActivationFunction"]
        self.mlp.weights = DATA["Weights"]
        self.mlp.biases = DATA["Biases"]

        return True
    
    def _showOutput(img, outputName = "Output", outputPath = "./"):
        cv2.imwrite(f"{outputPath if outputPath != './' else ''}{outputName}.png", np.array(dp.clip(img, 0, 255)).astype(np.uint8))

    def showOutputs(featuresMaps, prefix = "Output", outputPath = "./"):
        if isinstance(featuresMaps[0], int):
            CNNDebug._showOutput(featuresMaps, f"{prefix}", outputPath)
        else:
            for index, features in enumerate(featuresMaps):
                nombre = f"{prefix}{index}"

                progress = (index + 1) / len(featuresMaps)
                barra = ot.BarLoad(10)
                barra.showBar(progress, "Imprimiendo Imagenes: ", f" {index+1}/{len(featuresMaps)} {nombre} ")
                CNNDebug._showOutput(features, nombre, outputPath)


from zeroneura.core.ConvolutionalNeuralNetwork import CNNDebug as CNN
import zeroneura.utils.DataProcessing as dp
import unittest

kernel1 = [
    [-1, -2, -1],
    [ 0,  0,  0],
    [ 1,  2,  1]
]

kernel2 = [
    [-1, 0, 1],
    [-2, 0, 2],
    [-1, 0, 1]
]

kernel3 = [
    [0,  1,  2],
    [-1, 0,  1],
    [-2, -1, 0]
]

kernel4 = [
    [2, 1, 0],
    [1, 0, -1],
    [0, -1, -2]
]

kernel5 = [
    [0, -1, 0],
    [-1, 4, -1],
    [0, -1, 0]
]

kernel6 = [
    [-1, -1, -1],
    [-1, 8, -1],
    [-1, -1, -1]
]

kernel7 = [
    [1, -2, 1],
    [-2, 4, -2],
    [1, -2, 1]
]

kernel8 = [
    [-1, -1, -1],
    [0, 0, 0],
    [1, 1, 1]
]

kernel9 = [
    [-1, 0, 1],
    [-1, 0, 1],
    [-1, 0, 1]
]

kernel10 = [
    [0, 0, 0],
    [-1, 2, -1],
    [0, 0, 0]
]

kernels = [kernel1, kernel2, kernel3, kernel4, kernel5,kernel6, kernel7, kernel8, kernel9, kernel10]
frutas = ['Sandia', 'Manzana', 'Naranja', 'Fresa', 'FresaTR']
frutasI = [f"./tests/Inputs/CNN/{fruta}.png" for fruta in frutas]
frutasO = [1, 2, 3, 4, 4]
frutasO = [val/dp.getHigherTenMultiply(max(frutasO)) for val in frutasO]
inputPath = './tests/Outputs/CNN/'
inputName = 'TestCNN'

class TestCNN(unittest.TestCase):          
    def testCNNComplete(self):
        self.testCNNSave()
        self.testCNNLoad()

    def testCNNSave(self):        
        print("\n=== ENTRENANDO CNN ===\n")
        cnn = CNN([kernels], 5, 1)
        cnn.setMLPStructure([100, 64, 32, 1], learningRate=0.1)
        cnn.train(frutasI, frutasO, 200, True, inputPath)

        print("\n=== GUARDANDO DATOS ===\n")
        cnn.saveData(inputName, inputPath)

    def testCNNLoad(self):
        print("\n=== CARGANDO CNN Y PREDICIENDO ===\n")
        cnn1 = CNN()
        cnn1.loadData(inputName, inputPath)

        resultados = []
        for fruta in frutasI:
            print(f"\n>>> Predicción: {fruta}")
            res = cnn1.predict(fruta, inputPath + "/Test/", lambda x: round(x*10), True)
            resultados.append(res)

        print("\n=== RESULTADOS FINALES ===\n")
        for fruta, resultado, esperado in zip(frutas, resultados, frutasO):
            esperado *= 10
            procesado = dp.flatten(resultado)
            promedio = sum(procesado) / len(procesado)
            aciertos = sum(1 for x in procesado if round(x) == esperado)
            porcentaje_certeza = (aciertos / len(procesado)) * 100 if len(procesado) > 0 else 0

            print(f"Fruta: {fruta}")
            print(f"Esperado: {esperado}")
            print(f"Obtenido: {resultado}")
            print(f"\tPromedio: {promedio:.2f}")
            print(f"\tComparación: {round(promedio)} == {esperado} → {round(promedio) == esperado}")
            print(f"\tPorcentaje de Certeza: {porcentaje_certeza:.2f}%\n")

if __name__ == "__main__":
    unittest.main()


import zeroneura.utils.Tensors as ts
import zeroneura.utils.DataProcessing as dp
import cv2, numpy as np

def resize(inputPath, newSize, outputPath):
    img = cv2.imread(inputPath)
    resized = cv2.resize(img, newSize)

    cv2.imwrite(outputPath, resized)

def applyGrayScale(inputValue):
    img = cv2.imread(inputValue).tolist()
    alto, ancho = ts.getShapeTensor(img)[:2]

    for y in range(alto):
        for x in range(ancho):
            r, g, b = img[y][x]
            img[y][x] = int(0.299*r + 0.587*g + 0.114*b)
    
    return img

def normalizeGrayImg(img, maxValue = None):
    img = ts.applyLambdaToTensor(img, lambda x: max(0, min(255, x)))

    maxValue = max(img) if maxValue == None else maxValue
    maxValue = dp.getHigherTenMultiply(maxValue)
    img = ts.applyLambdaToTensor(img, lambda x: x/maxValue)

    return img
    
def applyPadding(img, padding = 0, value = 0):
    h, w = ts.getShapeTensor(img)[:2]
    padded = [[value]* (w + 2*padding) for _ in range(h + 2*padding)]
    
    for i in range(h):
        for j in range(w):
            padded[i + padding][j + padding] = img[i][j]

    return padded    

def showOutput(img, outputName = "Output", outputPath = "./"):
    cv2.imwrite(f"{outputPath if outputPath != './' else ''}{outputName}.png", np.array(dp.clip(img, 0, 255)).astype(np.uint8))


import zeroneura.utils.Tensors as ts
import cv2

def maxPooling(tensor, size = 2, stride = 2):
    height = len(tensor)
    width = len(tensor[0])
    pooled = []

    for i in range(0, height - size + 1, stride):
        row = []
        for j in range(0, width - size + 1, stride):
            submath = [tensor[i + dy][j + dx] for dy in range(size) for dx in range(size)]
            row.append(max(submath))
        pooled.append(row)        

    return pooled

def applyKernelFromPath(inputPath, kernel):
    imagen = ts.applyGrayScale(cv2.imread(inputPath).tolist())

    img = []
    alto, ancho = ts.getShapeTensor(imagen)[:2]
    for y in range(alto):
        row = []

        for x in range(ancho):
            submath = 0
            
            testWidth = x + 1 < ancho and x - 1 >= 0
            testHeight = y + 1 < alto and y - 1 >= 0
            if testWidth and testHeight:
                for i in range(len(kernel)):
                    for j in range(len(kernel)):
                        submath += kernel[i][j]*imagen[y + i - 1][x + j - 1]

            row.append(min(submath, 255))
        img.append(row)

    return img

def applyKernelWithStride(img, kernel, stride = 1):
        h, w = ts.getShapeTensor(img)[:2]
        kh, kw = ts.getShapeTensor(kernel)[:2]

        hOut, wOut = (h-kh)//stride + 1, (w-kw)//stride + 1
        output = [[0]*wOut for _ in range(hOut)]

        for y in range(0, h - kh + 1, stride):
            for x in range(0, w - kw + 1, stride):
                suma = 0

                for i in range(kh):
                    for j in range(kw):
                        suma += img[y+i][x+j] * kernel[i][j]

                output[y//stride][x//stride] = suma

        return output

def flatten(tensor):
    if isinstance(tensor, (int, float)):
        return [tensor]

    if not tensor:
        return []

    if isinstance(tensor[0], (int, float)):
        return list(tensor)

    flat = []
    for sub in tensor:
        flat += flatten(sub)
    return flat


def clip(tensor, mini, maxi):
    return ts.applyLambdaToTensor(tensor, lambda x: max(mini, min(maxi, x)))

def getHigherTenMultiply(val):
    r = val//10 + (1 if val%10 else 0)
    return r*10


import zeroneura.core.MultiLayerPerceptron as mlp
import zeroneura.utils.ActivationFunctions as af
import unittest

inputs = [
    [0, 0, 0, 0],
    [0, 0, 0, 1],
    [0, 0, 1, 0],
    [0, 0, 1, 1],
    [0, 1, 0, 0],
    [0, 1, 0, 1],
    [0, 1, 1, 0],
    [0, 1, 1, 1],
    [1, 0, 0, 0],
    [1, 0, 0, 1],
    [1, 0, 1, 0],
    [1, 0, 1, 1],
    [1, 1, 0, 0],
    [1, 1, 0, 1],
    [1, 1, 1, 0],
    [1, 1, 1, 1],
]

outputs = [
    [0, 0, 0, 1, 1, 1, 1, 1],
    [0, 1, 0, 1, 1, 1, 1, 0],
    [0, 1, 0, 1, 1, 1, 0, 1],
    [0, 1, 0, 1, 1, 1, 0, 0],
    [0, 1, 1, 0, 1, 0, 1, 1],
    [0, 1, 1, 0, 1, 0, 1, 0],
    [0, 1, 1, 0, 1, 0, 0, 1],
    [0, 1, 1, 0, 1, 0, 0, 0],
    [0, 1, 1, 0, 0, 1, 1, 1],
    [0, 1, 1, 0, 0, 1, 1, 0],
    [0, 1, 1, 0, 0, 1, 0, 1],
    [0, 1, 1, 0, 0, 1, 0, 0],
    [0, 1, 0, 1, 0, 0, 1, 1],
    [0, 1, 0, 1, 0, 0, 1, 0],
    [0, 1, 0, 1, 0, 0, 0, 1],
    [1, 1, 0, 1, 0, 0, 0, 0],
]

inputPath = './tests/Outputs/MLP/'
inputName = 'TestMLP'

class TestMLP(unittest.TestCase):
    def testMLP(self):
        print(f"\n{'-'*40}Prueba Unitaria Para MultiLayerPreptron{'-'*40}\n")

        mlp1 = mlp.MLP([len(inputs[0]), 8, len(outputs[0])], 0.1)
        mlp1.train(inputs, outputs, 20000)
        mlp1.printData()
        print("Test: ")
        mlp1.predictLambda(inputs, lambda r: round(r))

    def testMLPAF(self):
        print(f"\n{'-'*40}Prueba Unitaria Para MultiLayerPreptronActivationFunctions{'-'*40}\n")
        
        mlp2 = mlp.MLPAF([len(inputs[0]), 8, len(outputs[0])], 0.1, af.SIGMOID)
        mlp2.train(inputs, outputs, 2000)
        mlp2.printData()
        print("Test: ")
        mlp2.predictLambda(inputs, lambda r: round(r))
        
    def testMLPAFDebugSave(self):
        print(f"\n{'-'*40}Prueba Unitaria Para MultiLayerPreptronActivationFunctionsDebug{'-'*40}\n")

        mlp3 = mlp.MLPAFDebug([len(inputs[0]), 8, len(outputs[0])], 0.1, af.RELU)
        mlp3.train(inputs, outputs, 50000, True)
        mlp3.printData()
        
        mlp3.saveData(inputName, inputPath)
        
    def testMLPAFDebugLoad(self):
        print(f"\n{'-'*40}Prueba Unitaria Para MultiLayerPreptronActivationFunctionsDebug{'-'*40}\n")

        mlp4 = mlp.MLPAFDebug([len(inputs[0]), 8, len(outputs[0])], 0.1, af.RELU)        
        mlp4.loadData(inputName, inputPath)
        mlp4.printData()
        print("Test: ")
        mlp4.predict(inputs, lambda r: round(r))

if __name__ == "__main__":
    unittest.main()


import sys

class BarLoad:
    def __init__(self, length = 50, showPercent = True, filledBar = '█', unfilledBar = '-'):
        self.length = length
        self.filledBar = filledBar
        self.unfilledBar = unfilledBar
        self.showPercent = showPercent

    def showBar(self, progress, prefix = '', sufix = ''):
        progress = max(0, min(progress, 1))
        filledLength = int(self.length * progress)
        bar = self.filledBar * filledLength + self.unfilledBar * (self.length - filledLength)
        percentText = f"({progress * 100:.2f}%)" if self.showPercent else ''

        sys.stdout.write(f"\r{prefix}|{bar}|{sufix}{percentText}")
        sys.stdout.flush()


from zeroneura.core.Perceptron import Perceptron
import unittest

class TestMLP(unittest.TestCase):
    def testMLP(self):
        print(f"\n{'-'*40}Prueba Unitaria Para Preptron{'-'*40}\n")
        inputs = [
            [0, 0],
            [0, 1],
            [1, 0],
            [1, 1]
        ]

        outputs = [0, 0, 0, 1]

        perceptron = Perceptron([0.0, 0.0], 0.1, 0)
        perceptron.train(inputs, outputs, 100)
        perceptron.predict([[0, 0], [1, 1], [0, 1]], True)

if __name__ == "__main__":
    unittest.main()